import torchmetrics
import torch.nn as nn
import torch.optim as optim
import pytorch_lightning as pl
import segmentation_models_pytorch as smp


class MaterialMLP(pl.LightningModule):
    def __init__(self, n_in_channels, n_out_channels, lr=1e-4, wd=1e-5, opt_name='adam', ignore_index=0):
        super().__init__()

        self.lr = lr
        self.wd = wd
        self.opt_name = opt_name
        self.n_classes = n_out_channels
        self.in_channels = n_in_channels
        self.ignore_index = ignore_index

        # Build network.
        self.network = self._build_network()

        # Get metrics.
        metrics = torchmetrics.MetricCollection([
            torchmetrics.F1Score(task='multiclass',
                                 num_classes=self.n_classes,
                                 ignore_index=self.ignore_index,
                                 average='micro'),
            torchmetrics.JaccardIndex(task='multiclass',
                                      num_classes=self.n_classes,
                                      ignore_index=self.ignore_index,
                                      average='micro'),
            torchmetrics.Accuracy(task='multiclass',
                                  num_classes=self.n_classes,
                                  ignore_index=self.ignore_index,
                                  average='micro'),
        ])
        self.train_metrics = metrics.clone(prefix='train_')
        self.valid_metrics = metrics.clone(prefix='val_')
        self.test_metrics = metrics.clone(prefix='test_')

        # Get losses.
        self.loss_funcs = self._get_loss_funcs()

    def _build_network(self):
        network = nn.Sequential(nn.Linear(self.in_channels, self.in_channels), nn.BatchNorm1d(self.in_channels),
                                nn.PReLU(), nn.Linear(self.in_channels, self.in_channels),
                                nn.BatchNorm1d(self.in_channels), nn.PReLU(),
                                nn.Linear(self.in_channels, self.in_channels), nn.BatchNorm1d(self.in_channels),
                                nn.PReLU(), nn.Linear(self.in_channels, self.n_classes))
        return network

    def forward(self, batch):
        output = self.network(batch['pixel_data'].float())
        return output

    def _set_model_to_train(self):
        self.network.train()

    def _set_model_to_eval(self):
        self.network.eval()

    def training_step(self, batch, batch_idx):
        self._set_model_to_train()

        output = self.forward(batch)

        loss = self._compute_loss(output, batch['target'])

        # Track metrics.
        pred = output.argmax(dim=1)
        flat_pred, flat_target = pred.flatten(), batch['target'].flatten()
        metric_output = self.train_metrics(flat_pred, flat_target)

        # Log metrics and loss.
        metric_output['train_loss'] = loss
        self.log_dict(metric_output, prog_bar=True, on_step=True, on_epoch=True)

        return loss

    def validation_step(self, batch, batch_idx):
        self._set_model_to_eval()
        output = self.forward(batch)

        # Compute loss.
        loss = self._compute_loss(output, batch['target'])

        # Track metrics.
        pred = output.argmax(dim=1)
        flat_pred, flat_target = pred.flatten(), batch['target'].flatten()
        metric_output = self.valid_metrics(flat_pred, flat_target)
        self.valid_metrics.update(flat_pred, flat_target)

        # Log metrics and loss.
        metric_output['valid_loss'] = loss
        self.log_dict(metric_output, prog_bar=True, on_step=True, on_epoch=True)
        # self.log_dict({}, prog_bar=True, on_step=True, on_epoch=True)

        return metric_output, batch['target'].shape[0]

    def test_step(self, batch, batch_idx):
        self._set_model_to_eval()
        output = self.forward(batch)

        # Compute loss.
        loss = self._compute_loss(output, batch['target'])

        # Track metrics.
        pred = output.argmax(dim=1)
        flat_pred, flat_target = pred.flatten(), batch['target'].flatten()
        self.test_metrics.update(flat_pred, flat_target)

        # Log metrics and loss.
        self.log_dict({'test_loss': loss}, prog_bar=True, on_step=True, on_epoch=True)

    def configure_optimizers(self):
        if self.opt_name == 'adam':
            optimizer = optim.Adam(self.parameters(), lr=self.lr, weight_decay=self.wd)
        else:
            raise NotImplementedError(f'No implementation for optimizer of name: {self.opt_name}')
        return optimizer

    def validation_epoch_end(self, validation_step_outputs):
        if len(validation_step_outputs) == 0:
            self.test_f1_score = 0
            self.test_iou = 0
            self.test_acc = 0
        else:
            metric_output = self.valid_metrics.compute()
            self.log_dict(metric_output)

    def test_epoch_end(self, test_step_outputs) -> None:
        if len(test_step_outputs) == 0:
            pass
        else:
            metric_output = self.test_metrics.compute()
            self.log_dict(metric_output)

            self.f1_score = metric_output['test_F1Score'].item()
            self.acc = metric_output['test_Accuracy'].item()
            self.iou = metric_output['test_JaccardIndex'].item()

    def _get_loss_funcs(self):
        losses = {}
        losses['focal'] = smp.losses.FocalLoss(mode='multiclass', ignore_index=self.ignore_index)
        losses['dice'] = smp.losses.DiceLoss(mode='multiclass', ignore_index=self.ignore_index)
        losses['jaccard'] = smp.losses.JaccardLoss(mode='multiclass')
        losses['ce'] = nn.CrossEntropyLoss(ignore_index=self.ignore_index)
        return losses

    def _compute_loss(self, output, target):
        sum_loss = 0
        for loss_name, loss_func in self.loss_funcs.items():
            sum_loss += loss_func(output, target)
        return sum_loss



import ubelt as ub
import os
import pandas as pd
import pygtrie
from watch.utils import slugify_ext
from watch.utils.util_stringalgo import shortest_unique_suffixes


def pandas_reorder_columns(df, columns):
    remain = df.columns.difference(columns)
    return df.reindex(columns=(columns + list(remain)))


def pandas_argmaxima(data, columns, k=1):
    """
    Finds the top K indexes for each column.

    Args:
        data : pandas data frame
        columns : columns to maximize
        k : number of top entries per column

    Returns:
        List: indexes into subset of data that are in the top k for any of the
            requested columns.

    Example:
        >>> from watch.mlops.aggregate import *  # NOQA
        >>> import numpy as np
        >>> data = pd.DataFrame({k: np.random.rand(10) for k in 'abcde'})
        >>> columns = ['b', 'd', 'e']
        >>> k = 1
        >>> top_indexes = pandas_argmaxima(data=data, columns=columns, k=k)
        >>> print(data.loc[top_indexes])
    """
    top_indexes = None
    for col in columns:
        ranked_data = data[col].sort_values(ascending=False)
        ranked_idxs = ranked_data.index
        if top_indexes is None:
            top_indexes = ranked_idxs[0:k]
        else:
            top_indexes = top_indexes.union(ranked_idxs[0:k], sort=False)
    return top_indexes


def pandas_suffix_columns(data, suffixes):
    """
    Return columns that end with this suffix
    """
    return [c for c in data.columns if any(c.endswith(s) for s in suffixes)]


def pandas_nan_eq(a, b):
    nan_flags1 = pd.isna(a)
    nan_flags2 = pd.isna(b)
    eq_flags = a == b
    both_nan = nan_flags1 & nan_flags2
    flags = eq_flags | both_nan
    return flags


def pandas_shorten_columns(summary_table, return_mapping=False):
    """
    Shorten column names
    """
    import ubelt as ub
    # fixme
    old_cols = summary_table.columns
    new_cols = shortest_unique_suffixes(old_cols, sep='.')
    mapping = ub.dzip(old_cols, new_cols)
    summary_table = summary_table.rename(columns=mapping)
    if return_mapping:
        return summary_table, mapping
    else:
        return summary_table


def pandas_condense_paths(colvals):
    """
    Condense a column of paths
    """
    is_valid = ~pd.isnull(colvals)
    valid_vals = colvals[is_valid].apply(os.fspath)
    unique_valid_vals = valid_vals.unique().tolist()
    unique_short_vals = shortest_unique_suffixes(unique_valid_vals, sep='/')
    new_vals = [p.split('.')[0] for p in unique_short_vals]
    mapper = ub.dzip(unique_valid_vals, new_vals)
    condensed = colvals.apply(lambda x: mapper.get(x, x))
    return condensed, mapper


def pandas_truncate_items(data, paths=False, max_length=16):
    """
    from watch.utils.util_pandas import pandas_truncate_items

    Args:
        data (pd.DataFrame): data frame to truncate

    Returns:
        Tuple[pd.DataFrame, Dict[str, str]]
    """
    def truncate(x):
        if not isinstance(x, (str, os.PathLike)):
            return x
        return slugify_ext.smart_truncate(str(x), max_length=max_length, trunc_loc=0,
                                          hash_len=4, head='', tail='')
    mappings = {}
    if len(data):
        # only check the first row to see if we want to truncate the columns or
        # not
        trunc_str_cols = set()
        trunc_path_cols = set()
        for _, check_row in data.iloc[0:10].iterrows():
            for k, v in check_row.items():
                if paths:
                    # Check if probably a path or not
                    if isinstance(v, os.PathLike):
                        trunc_path_cols.add(k)
                        continue
                    elif isinstance(v, str) and '/' in v:
                        trunc_path_cols.add(k)
                        continue
                if isinstance(v, (str, os.PathLike)) and len(str(v)) > max_length:
                    trunc_str_cols.add(k)

        trunc_str_cols = list(ub.oset(data.columns) & trunc_str_cols)
        trunc_path_cols = list(ub.oset(data.columns) & trunc_path_cols)

        trunc_data = data.copy()
        trunc_data[trunc_str_cols] = trunc_data[trunc_str_cols].applymap(truncate)
        for c in trunc_data[trunc_str_cols]:
            v2 = pd.Categorical(data[c])
            data[c] = v2
            v1 = v2.map(truncate)
            mapping = list(zip(v1.categories, v2.categories))
            mappings[c] = mapping

        for c in trunc_path_cols:
            colvals = trunc_data[c]
            condensed, mapping = pandas_condense_paths(colvals)
            trunc_data[c] = condensed
            mappings[c] = mapping
    else:
        mapping = {}
        trunc_data = data
    return trunc_data, mappings


class DotDictDataFrame(pd.DataFrame):
    """
    A proof-of-concept wrapper around pandas that lets us walk down the nested
    structure a little easier.

    The API is a bit weird, and the caches are not invalidated if any column
    changes, but it does a reasonable job otherwise.

    Is there another library out there that does this?

    SeeAlso:
        DotDict

    Example:
        >>> from watch.utils.util_pandas import *  # NOQA
        >>> rows = [
        >>>     {'node1.id': 1, 'node2.id': 2, 'node1.metrics.ap': 0.5, 'node2.metrics.ap': 0.8},
        >>>     {'node1.id': 1, 'node2.id': 2, 'node1.metrics.ap': 0.5, 'node2.metrics.ap': 0.8},
        >>>     {'node1.id': 1, 'node2.id': 2, 'node1.metrics.ap': 0.5, 'node2.metrics.ap': 0.8},
        >>>     {'node1.id': 1, 'node2.id': 2, 'node1.metrics.ap': 0.5, 'node2.metrics.ap': 0.8},
        >>> ]
        >>> self = DotDictDataFrame(rows)
        >>> # Test prefix lookup
        >>> assert set(self['node1'].columns) == {'node1.id', 'node1.metrics.ap'}
        >>> # Test suffix lookup
        >>> assert set(self['id'].columns) == {'node1.id', 'node2.id'}
        >>> # Test mid-node lookup
        >>> assert set(self['metrics'].columns) == {'node1.metrics.ap', 'node2.metrics.ap'}
        >>> # Test single lookup
        >>> assert set(self[['node1.id']].columns) == {'node1.id'}
        >>> # Test glob
        >>> assert set(self.find_columns('*metri*')) == {'node1.metrics.ap', 'node2.metrics.ap'}
    """

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.__dict__['_trie_cache'] = {}

    def _clear_column_caches(self):
        self._trie_cache = {}

    @property
    def _column_prefix_trie(self):
        # TODO: cache the trie correctly
        if self._trie_cache.get('prefix_trie', None) is None:
            _trie_data = ub.dzip(self.columns, self.columns)
            _trie = pygtrie.StringTrie(_trie_data, separator='.')
            self._trie_cache['prefix_trie'] = _trie
        return self._trie_cache['prefix_trie']

    @property
    def _column_suffix_trie(self):
        if self._trie_cache.get('suffix_trie', None) is None:
            reversed_columns = ['.'.join(col.split('.')[::-1])
                                for col in self.columns]
            _trie_data = ub.dzip(reversed_columns, reversed_columns)
            _trie = pygtrie.StringTrie(_trie_data, separator='.')
            self._trie_cache['suffix_trie'] = _trie
        return self._trie_cache['suffix_trie']

    @property
    def _column_node_groups(self):
        if self._trie_cache.get('node_groups', None) is None:
            paths = [col.split('.') for col in self.columns]
            lut = ub.ddict(list)
            for path in paths:
                col = '.'.join(path)
                for part in path:
                    lut[part].append(col)
            self._trie_cache['node_groups'] = lut
        return self._trie_cache['node_groups']

    @property
    def nested_columns(self):
        from watch.utils.util_dotdict import dotkeys_to_nested
        return dotkeys_to_nested(self.columns)

    def find_column(self, col):
        result = self.query_column(col)
        if len(result) == 0:
            raise KeyError
        elif len(result) > 1:
            raise RuntimeError
        return list(result)[0]

    def query_column(self, col):
        # Might be better to do a globby sort of pattern
        parts = col.split('.')
        return ub.oset.intersection(*[self._column_node_groups[p] for p in parts])
        # try:
        #     candiates.update(self._column_prefix_trie.values(col))
        # except KeyError:
        #     ...
        # try:
        #     candiates.update(self._column_suffix_trie.values(col))
        # except KeyError:
        #     ...
        # return candiates

    def lookup_suffix_columns(self, col):
        return self._column_suffix_trie.values(col)

    def lookup_prefix_columns(self, col):
        return self._column_prefix_trie.values(col)

    def find_columns(self, pat, hint='glob'):
        from watch.utils import util_pattern
        pat = util_pattern.Pattern.coerce(pat, hint=hint)
        found = [c for c in self.columns if pat.match(c)]
        return found

    def subframe(self, key, drop_prefix=True):
        """
        Given a prefix key, return the subet columns that match it with the
        stripped prefix.
        """
        lookup_keys = []
        new_keys = []
        key_parts = key.split('.')
        nlevel = len(key_parts)
        for c in self.columns:
            path = c.split('.')
            if path[0:nlevel] == key_parts:
                lookup_keys.append(c)
                if drop_prefix:
                    new_keys.append('.'.join(path[nlevel:]))
        new = self.loc[:, lookup_keys]
        if drop_prefix:
            new.rename(ub.dzip(lookup_keys, new_keys), inplace=True, axis=1)
        return new

    def __getitem__(self, cols):
        if isinstance(cols, str):
            if cols not in self.columns:
                cols = self.query_column(cols)
                if not cols:
                    print(f'Available columns={self.columns}')
                    raise KeyError
        elif isinstance(cols, list):
            cols = list(ub.flatten([self.query_column(c) for c in cols]))
        return super().__getitem__(cols)


def pandas_add_prefix(data, prefix):
    return data.add_prefix(prefix)
    # mapper = {c: prefix + c for c in data.columns}
    # return data.rename(mapper, axis=1)



# Old plots that we might remove entirely soon


# def plot_ta1_vs_l1(merged_df, human_mapping, iarpa_metric_lut, pixel_metric_lut, common_plotkw, dpath, fnum=None):
#     import kwplot
#     sns = kwplot.autosns()

#     if fnum is None:
#         fnum = kwplot.next_fnum()

#     fnum = 0
#     expt_group = dict(list(merged_df.groupby(['test_dset', 'type'])))
#     k1 = ('Aligned-Drop3-TA1-2022-03-10', 'eval_trk+pxl')
#     k2 = ('Aligned-Drop3-L1', 'eval_trk+pxl')
#     plot_name = 'ta1_vs_l1'
#     param = 'Processing'
#     plotkw = ub.udict({
#         'x': 'salient_AP',
#         'y': 'BAS_F1',
#         'hue': param,
#         # 'hue': 'sensorchan',
#         **common_plotkw,
#         # 'hue': 'trk_use_viterbi',
#         # 'style': 'trk_thresh',
#         # 'size': 'trk_thresh',
#         # 'hue': 'pred_cfg',
#         # 'hue': 'expt',
#     })
#     x = plotkw['x']
#     y = plotkw['y']
#     plotkw.pop('style', None)

#     plot_dpath = (dpath / plot_name).ensuredir()

#     from watch.utils import result_analysis
#     all_param_keys = ub.oset.union(trk_param_keys, pred_param_keys,
#                                    fit_param_keys, act_param_keys)

#     all_param_keys = {
#         'trk_thresh',
#         # 'trk_morph_kernel',
#         'trk_agg_fn', 'trk_thresh_hysteresis', 'trk_moving_window_size',
#         'pred_tta_fliprot', 'pred_tta_time', 'pred_chip_overlap',
#         # 'sensorchan',
#         # 'time_steps',
#         # 'chip_size',
#         # 'chip_overlap',
#         # 'arch_name',
#         # 'optimizer', 'time_sampling', 'time_span', 'true_multimodal',
#         # 'accumulate_grad_batches', 'modulate_class_weights', 'tokenizer',
#         # 'use_grid_positives', 'use_cloudmask', 'upweight_centers',
#         # 'temporal_dropout', 'stream_channels', 'saliency_loss', 'class_loss',
#         # 'init', 'learning_rate', 'decoder',
#         'trk_use_viterbi'
#     }

#     data1 = expt_group[k1]
#     data2 = expt_group[k2]
#     data = pd.concat([data1, data2])
#     data = data[~data['has_teamfeat']]

#     if 0:
#         # Reduce to comparable groups according to the abalate criterion
#         results_list = []
#         for row in data.to_dict('records'):
#             params = ub.dict_isect(row, {'Processing', *all_param_keys})
#             metrics = ub.dict_isect(row, {x, y})
#             result = result_analysis.Result(None, params, metrics)
#             results_list.append(result)
#         self = analysis = result_analysis.ResultAnalysis(
#             results_list, default_objective='max', metrics={'BAS_F1'})
#         comparable_data = []
#         for group in self.abalation_groups('Processing'):
#             print(len(group))
#             if len(group['Processing'].unique()) > 1:
#                 comparable_data.append(group)
#         comparable = pd.concat(comparable_data)
#         data = data.iloc[comparable.index]

#     if 0:
#         # Remove duplicates for clarity
#         rows = []
#         for model, group in data.groupby('model'):
#             if len(group) > 1:
#                 # group.pred_cfg.value_counts()
#                 # group.trk_cfg.value_counts()
#                 idx = group[y].argmax()
#                 row = group.iloc[idx]
#             else:
#                 row = group.iloc[0]
#             rows.append(row)
#         data = pd.DataFrame(rows)

#     results_list = []
#     for row in data.to_dict('records'):
#         # params = ub.dict_isect(row, {'Processing', *all_param_keys})
#         params = ub.dict_isect(row, {'Processing'})
#         metrics = ub.dict_isect(row, {x, y})
#         result = result_analysis.Result(None, params, metrics)
#         results_list.append(result)
#     analysis = result_analysis.ResultAnalysis(
#         results_list, default_objective='max', metrics={'BAS_F1', 'salient_AP'})

#     try:
#         analysis.run()
#     except TypeError:
#         raise

#     # kitware_green = '#3caf49'
#     # kitware_blue = '#006ab6'
#     kitware_green = '#3EAE2B'
#     kitware_blue = '#0068C7'

#     self = analysis
#     conclusions = analysis.conclusions()

#     fig = kwplot.figure(fnum=fnum, doclf=True)
#     ax = fig.gca()
#     palette = {
#         'L1': kitware_blue,
#         'TA1': kitware_green,
#     }
#     ax = humanized_scatterplot(human_mapping, data=data, ax=ax, legend=True, palette=palette, **plotkw)
#     # nice_type = human_mapping.get(type, type)
#     # ax.set_title('TA1 vs L1' + '\n' + '\n'.join(conclusions))
#     ax.set_title('TA1 vs L1')
#     fname = f'{plot_name}_scatter.png'
#     fpath = plot_dpath / fname
#     fig.set_size_inches(np.array([6.4, 4.8]) * 1.4)
#     fig.tight_layout()
#     fig.savefig(fpath)

#     bas_conclusion = '\n'.join([c for c in conclusions if 'BAS_F1' in c])

#     fnum = fnum + 1
#     fig = kwplot.figure(fnum=fnum, doclf=True)
#     ax = fig.gca()
#     ax.set_title('BAS scores: TA1 vs L1')
#     sns.violinplot(data=merged_df, x='Processing', y='BAS_F1', palette=palette)
#     ax.set_title('TA1 vs L1' + '\n' + bas_conclusion)
#     fname = f'{plot_name}_violin.png'
#     fpath = plot_dpath / fname
#     fig.set_size_inches(np.array([6.4, 4.8]) * 1.0)
#     fig.tight_layout()
#     fig.savefig(fpath)
#     cropwhite_ondisk(fpath)

#     fnum = fnum + 1
#     fig = kwplot.figure(fnum=fnum, doclf=True)
#     ax = fig.gca()
#     sns.boxplot(data=merged_df, x='Processing', y='BAS_F1', palette=palette)
#     ax.set_title('TA1 vs L1' + '\n' + bas_conclusion)
#     fname = f'{plot_name}_boxwhisker.png'
#     fpath = plot_dpath / fname
#     fig.set_size_inches(np.array([6.4, 4.8]) * 1.0)
#     fig.tight_layout()
#     fig.savefig(fpath)
#     cropwhite_ondisk(fpath)

#     # ax.set_title('TA1 vs L1' + '\n' + '\n'.join(conclusions))


# def plot_viterbii_analysis(merged_df, human_mapping, iarpa_metric_lut, pixel_metric_lut, common_plotkw):
#     import kwplot
#     expt_group = dict(list(merged_df.groupby(['test_dset', 'type'])))
#     from watch.utils import result_analysis

#     nan_defaults = {
#         'modulate_class_weights': '',
#         'trk_agg_fn': 'probs',
#         'pred_tta_fliprot': 0,
#         'pred_tta_time': 0,
#         'pred_chip_overlap': 0.3,
#         'decoder': 'mlp',
#         'trk_morph_kernel': 3,
#         'stream_channels': 8,
#         'trk_thresh': 0.2,
#         'trk_use_viterbi': 0,
#         'trk_thresh_hysteresis': None,
#         'trk_moving_window_size': None,
#         'use_cloudmask': 0,
#     }

#     merged_df.loc[merged_df['use_cloudmask'].isnull(), 'use_cloudmask'] = 0

#     code_type = ('Cropped-Drop3-TA1-2022-03-10', 'eval_act+pxl')
#     test_dset, type = code_type
#     group = expt_group[code_type]

#     iarpa_metric = iarpa_metric_lut[type]
#     pixel_metric = pixel_metric_lut[type]
#     # metric = pixel_metric
#     metric = iarpa_metric

#     results_list = []
#     for row in group.to_dict('records'):
#         metric_val = row[metric]
#         if math.isnan(metric_val):
#             continue
#         metrics = {
#             metric: metric_val,
#         }
#         params = ub.dict_isect(row, fit_param_keys + pred_param_keys + trk_param_keys + act_param_keys)
#         params['modulate_class_weights']

#         for k, v in params.items():
#             if isinstance(v, float) and math.isnan(v):
#                 if k == 'sensorchan':
#                     params['sensorchan'] = params['channels']
#                 else:
#                     params[k] = nan_defaults[k]

#         result = result_analysis.Result(None, params, metrics)
#         results_list.append(result)

#     ignore_params = {'bad_channels'}
#     ignore_metrics = {}
#     abalation_orders = {1}
#     analysis = result_analysis.ResultAnalysis(
#         results_list, ignore_params=ignore_params,
#         # metrics=['coi_mAPUC', 'coi_APUC'],
#         # metrics=['salient_AP'],
#         # metrics=['coi_mAP', 'salient_AP'],
#         metrics=[metric],
#         metric_objectives={
#             'salient_AP': 'max',
#             'coi_mAP': 'max',
#             'mean_f1': 'max',
#         },
#         ignore_metrics=ignore_metrics,
#         abalation_orders=abalation_orders
#     )
#     # try:
#     #     analysis.run()
#     # except TypeError:
#     #     raise

#     param = 'trk_use_viterbi'
#     scored_obs = analysis.abalate(param)
#     ab_rows = []
#     pts1 = []
#     pts2 = []
#     for obs in scored_obs:
#         ab_row = obs.melt(['trk_use_viterbi']).pivot(['variable'], ['trk_use_viterbi'], 'value').reset_index(drop=True)
#         if (~ab_row.isnull()).values.sum() > 1:

#             # hack
#             obspt = obs.copy()
#             obspt[param] = obs[param].astype(bool).astype(int)
#             pts1.append(obspt.values[0])
#             pts2.append(obspt.values[1])
#             ab_rows.append(ab_row)
#     # ab_df = pd.concat(ab_rows).reset_index(drop=True)
#     # obs[param]

#     plt = kwplot.autoplt()
#     ax = plt.gca()
#     kwplot.draw_line_segments(pts1, pts2)
#     # ticks = ax.set_xticks([0, 1], ['0', 'v6,v1'])
#     ax.set_ylabel(metric)
#     ax.set_xlabel(param)
#     ax.set_title('Viterbi A/B tests')

#     return

#     # ab_df
#     # .reset_index()
#     # ab_melt_df = ab_df.melt(id_vars=['index'], value_vars=ab_row.columns)
#     # sns = kwplot.autosns()
#     # sns.lineplot(data=ab_melt_df, x=param, y='value')

#     import seaborn as sns
#     kwplot.figure(fnum=1000)
#     sns.violinplot(data=merged_df, x='temporal_dropout', y=pixel_metric)

#     # TODO: translate to chip_dims
#     # kwplot.figure(fnum=1001)
#     # sns.violinplot(data=merged_df, x='chip_size', y=pixel_metric)

#     kwplot.figure(fnum=1002, doclf=True)
#     sns.violinplot(data=merged_df, x='time_steps', y=pixel_metric)

#     kwplot.figure(fnum=1003, doclf=True)
#     sns.violinplot(data=merged_df, x='trk_use_viterbi', y=pixel_metric)

#     kwplot.figure(fnum=1004, doclf=True)
#     ax = sns.violinplot(data=group, x='sensorchan', y=pixel_metric)
#     for xtick in ax.get_xticklabels():
#         xtick.set_rotation(90)

#     kwplot.figure(fnum=3)
#     sns.violinplot(data=merged_df, x='saliency_loss', y=pixel_metric)

#     group[heuristics.fit_param_keys]
#     group[heuristics.fit_param_keys]



import torch
from torch.utils import data
import os
import os.path as osp


def main(data_dir):
    data_folder = os.listdir(data_dir)

    train_folders, val_folders = data.random_split(
        data_folder, [50, 10], generator=torch.Generator().manual_seed(12))

    train_images = []
    train_building_masks = []
    for x in train_folders:
        for y in os.listdir(osp.join(data_dir, x, 'images')):
            train_images.append(osp.join(data_dir, x, 'images', y))
        for z in os.listdir(osp.join(data_dir, x, 'building_masks')):
            train_building_masks.append(osp.join(data_dir, x, 'building_masks', z))

    train_images = sorted(train_images)
    train_building_masks = sorted(train_building_masks)

    val_images = []
    val_building_masks = []
    for x in val_folders:
        for y in os.listdir(osp.join(data_dir, x, 'images')):
            val_images.append(osp.join(data_dir, x, 'images', y))
        for z in os.listdir(osp.join(data_dir, x, 'building_masks')):
            val_building_masks.append(osp.join(data_dir, x, 'building_masks', z))

    val_images = sorted(val_images)
    val_building_masks = sorted(val_building_masks)

    with open('./spacenet/data/splits_unmasked/spacenet7_train_images.txt', 'w') as file:
        for item in train_images:
            file.write('%s\n' % item)

    with open('./spacenet/data/splits_unmasked/spacenet7_train_building_masks.txt', 'w') as file:
        for item in train_building_masks:
            file.write('%s\n' % item)

    with open('./spacenet/data/splits_unmasked/spacenet7_val_images.txt', 'w') as file:
        for item in val_images:
            file.write('%s\n' % item)

    with open('./spacenet/data/splits_unmasked/spacenet7_val_building_masks.txt', 'w') as file:
        for item in val_building_masks:
            file.write('%s\n' % item)


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_dir', type=str, default='/localdisk0/SCRATCH/watch/SpaceNet/7/train/')
    args = parser.parse_args()

    main(args.data_dir)



from collections import deque
import itertools as it


def consume(iterator, n=None):
    """
    Consume n items from an iterator and discard them.

    Args:
        iterator (Iterable): an iterator to consume
        n (int | None): number of items to consume (or consume all if None)

    References:
        https://stackoverflow.com/questions/50937966/fastest-most-pythonic-way-to-consume-an-iterator
        https://docs.python.org/3/library/itertools.html#itertools-recipes

    Benchmark:
        >>> from watch.utils.util_iter import *  # NOQA
        >>> import timerit
        >>> ti = timerit.Timerit(100, bestof=10, verbose=2)
        >>> #
        >>> def make_iterator():
        >>>     return iter(range(100000))
        >>> #
        >>> for timer in ti.reset('our-consume'):
        >>>     iterator = make_iterator()
        >>>     with timer:
        >>>         consume(iterator)
        >>> #
        >>> for timer in ti.reset('list'):
        >>>     iterator = make_iterator()
        >>>     with timer:
        >>>         list(iterator)
        >>> #
        >>> for timer in ti.reset('consume-100'):
        >>>     iterator = make_iterator()
        >>>     with timer:
        >>>         consume(iterator, n=100)
        >>> #
        >>> for timer in ti.reset('list-100'):
        >>>     iterator = make_iterator()
        >>>     with timer:
        >>>         list(zip(iterator, range(100)))
    """
    if n is None:
        # consume the entire iterator
        deque(iterator, maxlen=0)
    else:
        # advance to the empty slice starting at position n
        next(it.islice(iterator, n, n), None)


def chunks(items, nchunks):
    """

    Note:
        ubelt.chunks does not handle this case (yet)

    TODO:
        - [ ] Fix ubelt.chunks to handle this remainder case

    Example:
        >>> from watch.utils.util_iter import *  # NOQA
        >>> items = list(range(11))
        >>> nchunks = 4
        >>> list(chunks(items, nchunks))
    """
    if nchunks == 0:
        return
    num_items = len(items)
    # Basic chunksize
    chunksize = num_items // nchunks
    # How should the remainder be distributed?
    # Evenly at the start or at the end? (for now do it at the start)
    remainder = num_items % nchunks

    # Build an iterator that describes how big each chunk will be
    chunksize_iter = it.chain(
        it.repeat(chunksize + 1, remainder),
        it.repeat(chunksize, nchunks - remainder)
    )
    iterator = iter(items)
    for _chunksize in chunksize_iter:
        yield list(it.islice(iterator, _chunksize))



"""
TODO:
    - [ ] This needs to be cleaned up. Python logging is way too global for my
    taste. I would prefer something more instance level, but that will require
    some thought. The netharn.FitHarn has a way of accomplishing this that I
    reasonably like. That might be worth generalizing and porting here.  But
    for now, this will remain somewhat ad-hoc.
"""
import logging.config
import warnings
from logging import CRITICAL, ERROR, WARNING, INFO, DEBUG


def setup_logging(verbose=1):
    """
    Define logging level

    Args:
        verbose (int):
            Accepted values:
                * 0: no logging
                * 1: INFO level
                * 2: DEBUG level
    """

    log_med = "%(asctime)s-15s %(name)-32s [%(levelname)-8s] %(message)s"
    log_large = "%(asctime)s-15s %(name)-32s [%(levelname)-8s] "
    log_large += "(%(module)-17s) %(message)s"

    log_config = {}

    if verbose > 2:
        warnings.warn(f'watch util_logging only accepts a maximum verbosity of 2. Reconfiguring {verbose} to 2.')
        verbose = 2

    if verbose == 0:
        log_config = {
            "version": 1,
            "disable_existing_loggers": False,
            "handlers": {
                "null": {"level": "DEBUG", "class": "logging.NullHandler"}
            },
            "loggers": {
                "watchlog": {
                    "handlers": ["null"],
                    "propagate": True,
                    "level": "INFO"
                }
            },
        }
    elif verbose == 1:
        log_config = {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "standard": {
                    "format": log_med
                }
            },
            "handlers": {
                "console": {
                    "level": "DEBUG",
                    "class": "logging.StreamHandler",
                    "formatter": "standard",
                }
            },
            "loggers": {
                "watchlog": {
                    "handlers": ["console"],
                    "propagate": True,
                    "level": "INFO",
                }
            },
        }
    elif verbose == 2:
        log_config = {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "verbose": {
                    "format": log_large
                }
            },
            "handlers": {
                "console": {
                    "level": "DEBUG",
                    "class": "logging.StreamHandler",
                    "formatter": "verbose",
                }
            },
            "loggers": {
                "watchlog": {
                    "handlers": ["console"],
                    "propagate": True,
                    "level": "DEBUG",
                }
            },
        }
    else:
        raise ValueError("'verbose' must be one of: 0, 1, 2")
    return log_config


def get_logger(verbose=1):
    logcfg = setup_logging(verbose)
    logging.config.dictConfig(logcfg)
    logger = logging.getLogger('watchlog')
    return logger


class PrintLogger:
    """
    Simple print-based logger that duck-types the logging.Logger class and
    "simply works" without configuration.

    Example:
        >>> from watch.utils.util_logging import *  # NOQA
        >>> logger = PrintLogger(level=logging.INFO)
        >>> logger.info('hello')
        >>> logger.debug('world')
        hello
        >>> logger = PrintLogger(level=logging.DEBUG)
        >>> logger.info('hello')
        >>> logger.debug('world')
        hello
        world
    """

    def __init__(self, name='<print-logger>', level=None, verbose=1):
        self.name = name
        if level is None:
            if verbose == 0:
                level = CRITICAL
            elif verbose == 1:
                level = INFO
            elif verbose == 2:
                level = DEBUG
        self.level = level
        self.parent = None
        self.propagate = True
        self.handlers = []
        self.disabled = False
        self._cache = {}

    def setLevel(self, level):
        raise NotImplementedError

    def debug(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'DEBUG'.
        """
        if self.isEnabledFor(DEBUG):
            self._log(DEBUG, msg, args, **kwargs)

    def info(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'INFO'.
        """
        if self.isEnabledFor(INFO):
            self._log(INFO, msg, args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'WARNING'.
        """
        if self.isEnabledFor(WARNING):
            self._log(WARNING, msg, args, **kwargs)

    def error(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'ERROR'.
        """
        if self.isEnabledFor(ERROR):
            self._log(ERROR, msg, args, **kwargs)

    def exception(self, msg, *args, exc_info=True, **kwargs):
        """
        Convenience method for logging an ERROR with exception information.
        """
        self.error(msg, *args, exc_info=exc_info, **kwargs)

    def critical(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'CRITICAL'.
        """
        if self.isEnabledFor(CRITICAL):
            self._log(CRITICAL, msg, args, **kwargs)

    def log(self, level, msg, *args, **kwargs):
        """
        Log 'msg % args' with the integer severity 'level'.
        """
        if not isinstance(level, int):
            raise TypeError("level must be an integer")
        if self.isEnabledFor(level):
            self._log(level, msg, args, **kwargs)

    def findCaller(self, stack_info=False, stacklevel=1):
        """
        Find the stack frame of the caller so that we can note the source
        file name, line number and function name.
        """
        raise NotImplementedError

    def makeRecord(self, name, level, fn, lno, msg, args, exc_info,
                   func=None, extra=None, sinfo=None):
        """
        A factory method which can be overridden in subclasses to create
        specialized LogRecords.
        """
        return msg

    def filter(self, record):
        return True

    def _log(self, level, msg, args, exc_info=None, extra=None, stack_info=False,
             stacklevel=1):
        """
        Low-level logging routine which creates a LogRecord and then calls
        all the handlers of this logger to handle the record.
        """
        import sys
        sinfo = None
        fn, lno, func = "(unknown file)", 0, "(unknown function)"
        if exc_info:
            if isinstance(exc_info, BaseException):
                exc_info = (type(exc_info), exc_info, exc_info.__traceback__)
            elif not isinstance(exc_info, tuple):
                exc_info = sys.exc_info()
        record = self.makeRecord(self.name, level, fn, lno, msg, args,
                                 exc_info, func, extra, sinfo)
        self.handle(record)

    def handle(self, record):
        """
        Call the handlers for the specified record.

        This method is used for unpickled records received from a socket, as
        well as those created locally. Logger-level filtering is applied.
        """
        if (not self.disabled) and self.filter(record):
            self.callHandlers(record)

    def addHandler(self, hdlr):
        """
        Add the specified handler to this logger.
        """
        raise NotImplementedError

    def removeHandler(self, hdlr):
        """
        Remove the specified handler from this logger.
        """
        raise NotImplementedError

    def hasHandlers(self):
        """
        See if this logger has any handlers configured.

        Loop through all handlers for this logger and its parents in the
        logger hierarchy. Return True if a handler was found, else False.
        Stop searching up the hierarchy whenever a logger with the "propagate"
        attribute set to zero is found - that will be the last logger which
        is checked for the existence of handlers.
        """
        raise NotImplementedError

    def callHandlers(self, record):
        print(record)
        # raise NotImplementedError

    def getEffectiveLevel(self):
        """
        Get the effective level for this logger.
        """
        return self.level

    def isEnabledFor(self, level):
        """
        Is this logger enabled for level 'level'?
        """
        if self.disabled:
            return False
        return level >= self.level

    def getChild(self, suffix):
        raise NotImplementedError

    def __repr__(self):
        level = logging.getLevelName(self.getEffectiveLevel())
        return '<%s %s (%s)>' % (self.__class__.__name__, self.name, level)

    # def __reduce__(self):
    #     raise NotImplementedError



import os
import pickle
import hashlib
import multiprocessing
from datetime import datetime

import kwcoco
import numpy as np
from einops import rearrange
from omegaconf import OmegaConf

from watch.tasks.rutgers_material_seg_v2.datasets.base_dataset import BaseDataset
from watch.tasks.rutgers_material_seg_v2.utils.util_dataset import filter_image_ids_by_sensor
from watch.tasks.rutgers_material_seg_v2.utils.util_misc import generate_video_slice_object, get_crop_slices


def process_example(image_id, image_slice, kwcoco_dset, channels):
    examples = []
    # Get datetime for image.
    image_name = kwcoco_dset.index.imgs[image_id]['name']
    year = int(image_name.split('_')[1][:4])
    month = int(image_name.split('_')[1][4:6])
    day = int(image_name.split('_')[1][6:8])
    dt = datetime(year, month, day)

    pct_load_image = 0.0
    if np.random.rand() < pct_load_image:
        # Load and scale image.
        # image = self._load_norm_image(image_id)
        img_delayed = kwcoco_dset.delayed_load(image_id, channels=channels)
        image = img_delayed.finalize()  # [height, width, channels]
        image = image.transpose(2, 0, 1)  # [channels, height, width]
        image = image / 2**16

        img_h, img_w = image.shape[-2:]

        # Randomly sample pixels for computing mean and std.
        ## TODO: Make percentage of pixels taken adjustable.
        pct = 0.001
        n_pixels = img_h * img_w
        n_sampled_pixels = int(pct * n_pixels)

        X = np.random.randint(0, img_h, size=n_sampled_pixels)
        Y = np.random.randint(0, img_w, size=n_sampled_pixels)

        sampled_pixels = image[:, X, Y]
    else:
        img_h = kwcoco_dset.index.imgs[image_id]['height']
        img_w = kwcoco_dset.index.imgs[image_id]['width']
        sampled_pixels = None

    # Get viable crop slices.
    crop_slices = get_crop_slices(img_h, img_w, crop_height=image_slice.height, crop_width=image_slice.width)

    for crop_slice in crop_slices:
        example = {}
        example['image_id'] = image_id
        example['crop_slice'] = crop_slice
        example['og_height'] = img_h
        example['og_width'] = img_w
        example['datetime'] = dt
        examples.append(example)

    return examples, sampled_pixels


class ImageDataset(BaseDataset):
    def __init__(self,
                 kwcoco_path,
                 split,
                 image_slice_cfg=None,
                 feature_type='pixel',
                 n_clusters=5,
                 region_shape='square',
                 square_size=5,
                 channels='ALL',
                 seed_num=0,
                 sensor='S2',
                 image_scale_mode='direct',
                 image_norm_mode=None,
                 overwrite_stats=None):

        # Create private variables.
        self.channels = channels
        self.n_clusters = n_clusters
        self.square_size = square_size
        self.feature_type = feature_type
        self.region_shape = region_shape
        self.image_norm_mode = image_norm_mode

        # Check image slice.
        if image_slice_cfg is None:
            raise ValueError('No Image Slice parameters given for ImageDataset.')
        else:
            self.image_slice = generate_video_slice_object(height=image_slice_cfg.height,
                                                           width=image_slice_cfg.width,
                                                           n_frames=image_slice_cfg.n_frames,
                                                           scale=image_slice_cfg.scale,
                                                           stride=image_slice_cfg.stride)

        # Call super function.
        super(ImageDataset, self).__init__(kwcoco_path,
                                           split,
                                           image_scale_mode=image_scale_mode,
                                           seed_num=seed_num,
                                           sensor=sensor)

        # Get examples.
        self._build_dataset()

        if overwrite_stats:
            self.mean = overwrite_stats['mean']
            self.std = overwrite_stats['std']

    def get_n_channels(self):
        # Update channels.
        # TODO: Fix this hack.
        if self.channels == 'ALL':
            print('WARNING: ALL channels does NOT include cloud channel or materials.')
            self.channels = 'red|green|blue|nir|swir16|swir22'

        n_channels = len(self.channels.split('|'))

        return n_channels

    def _create_cach_file_name(self):
        input_str = f'{self.split}_{self.feature_type}_{self.n_clusters}_{self.image_scale_mode}_{self.square_size}_{self.sensor}_{self.channels}'
        sha = hashlib.sha256()
        sha.update(input_str.encode())
        output_str = sha.hexdigest()

        return output_str + '.p'

    def _build_dataset(self):
        # Check to see if examples are already generated.
        self.dset_root = os.path.split(self.kwcoco_path)[0]
        cache_dir = os.path.join(self.dset_root, '_cache', 'image')
        os.makedirs(cache_dir, exist_ok=True)
        cache_file_name = self._create_cach_file_name()
        cache_path = os.path.join(cache_dir, cache_file_name)

        if os.path.exists(cache_path) is False:
            # Create examples.
            print(f'No cached examples found at: {cache_path}')

            # Load KWCOCO file.
            self.coco_dset = kwcoco.CocoDataset(self.kwcoco_path)

            # Get all image ids from sensor.

            ## Filter image IDs based on input sensor.
            image_ids = filter_image_ids_by_sensor(self.coco_dset, self.sensor)
            print(f'Number of image IDs after filtering by sensor ({self.sensor}): {len(image_ids)}')

            print(f'Multiprocessing on {multiprocessing.cpu_count()} cores: Gathering examples')
            self.examples, pixel_data = [], []
            # with multiprocessing.Pool() as pool:
            #     input_to_func = [(image_id, self.image_slice, self.coco_dset, self.channels) for image_id in image_ids]
            #     for (examples, ex_pixel_data) in pool.starmap(process_example,
            #                                                   tqdm(input_to_func, total=len(input_to_func))):
            #         self.examples.extend(examples)
            #         pixel_data.append(ex_pixel_data)
            for image_id in image_ids:
                examples, ex_pixel_data = process_example(image_id, self.image_slice, self.coco_dset, self.channels)
                self.examples.extend(examples)
                pixel_data.append(ex_pixel_data)

            # Combine pixel data.
            ## Filter Nones.
            # filtered_pixel_data = []
            # for data in pixel_data:
            #     if data is not None:
            #         filtered_pixel_data.append(data)
            # pixel_data = np.concatenate(filtered_pixel_data, axis=1)

            # Compute mean and std of this slice of dataset if split == 'train'.
            # mean, std = self._compute_dataset_stats(pixel_data)
            mean, std = None, None

            # Save examples and info needed to load from examples.
            cache_object = {'examples': self.examples, 'coco_dset': self.coco_dset, 'mean': mean, 'std': std}
            pickle.dump(cache_object, open(cache_path, 'wb'))
            print(f'Saved cached dataset to: {cache_path}')
        else:
            # Load cached examples.
            print(f'Loading examples from: {cache_path}')
            cache_object = pickle.load(open(cache_path, 'rb'))

        self.examples = cache_object['examples']
        self.coco_dset = cache_object['coco_dset']
        # self.mean, self.std = cache_object['mean'], cache_object['std']
        self.mean, self.std = None, None

    def __getitem__(self, index):
        example = self.examples[index]

        # Load image.
        crop = self._load_norm_image(example['image_id'], crop_slice=example['crop_slice'], channels=self.channels)

        # Buffer image.
        crop, buffer_mask = self._add_buffer_to_image(crop,
                                                      self.image_slice.height,
                                                      self.image_slice.width,
                                                      constant_value=-1)

        # Normalize data.
        # crop = (crop - self.mean[:, None, None]) / (self.std[:, None, None] + 1e-5)

        output = {}
        output['image'] = crop
        output['image_id'] = example['image_id']
        output['og_height'] = example['og_height']
        output['og_width'] = example['og_width']
        output['datetime'] = example['datetime']
        output['crop_slice'] = example['crop_slice']
        output['buffer_mask'] = buffer_mask
        return output

    def to_RGB(self, image, gamma=0.4):
        # image: [C, H, W]
        rgb = image[:3]
        rgb = np.clip(rgb**gamma, 0, 1)

        # rgb: [3, H, W]
        return rgb.transpose(1, 2, 0)


def test_get_pixel_dist(args, n_samples=200):
    # Build dataset.
    print('Building dataset ...')
    image_slice_dict = {
        'height': args.crop_size,
        'width': args.crop_size,
        'scale': 1,
        'stride': args.crop_size,
        'n_frames': None
    }
    image_slice_cfg = OmegaConf.create(image_slice_dict)
    dataset = ImageDataset(args.dataset_path, split=args.split, image_slice_cfg=image_slice_cfg)

    # Gather examples from dataset.
    print(f'\nCollecting {n_samples} samples from each dataset ...')
    indices = np.random.choice(list(range(dataset.__len__())), n_samples, replace=False)
    sample_data = []
    for index in indices:
        ex = dataset.__getitem__(index)
        image = ex['image']
        pixels = rearrange(image, 'c h w -> (h w) c')
        indices = np.where(pixels >= 0)
        subpixels = pixels[indices[0], :]
        sample_data.append(subpixels)
    sample_data = np.concatenate(sample_data, axis=0)

    print('\nPixel Distribution')
    print('-' * 20)
    print(f'Pixel dist: {sample_data.mean(axis=0)}')


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--dataset_path', type=str, help='Path to dataset file.')
    parser.add_argument('--split', type=str, default='valid')
    parser.add_argument('--crop_size', type=int, default=300)
    parser.add_argument('--feature_type', type=str, default='pixel')
    parser.add_argument('--test_type', type=str, default=None)
    parser.add_argument('--seed_num', type=int, default=0)
    args = parser.parse_args()

    # Set random seeds.
    import torch
    np.random.seed(args.seed_num)
    torch.random.manual_seed(args.seed_num)

    if args.test_type is None:
        image_slice_dict = {
            'height': args.crop_size,
            'width': args.crop_size,
            'scale': 1,
            'stride': args.crop_size,
            'n_frames': None
        }
        image_slice_cfg = OmegaConf.create(image_slice_dict)
        dataset = ImageDataset(args.dataset_path, split=args.split, image_slice_cfg=image_slice_cfg)
        dataset.__getitem__(0)

    elif args.test_type == 'get_pixel_dist_values':
        test_get_pixel_dist(args)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import torch
import torch.nn as nn


class MoCo(nn.Module):
    """
    Build a MoCo model with: a query encoder, a key encoder, and a queue
    https://arxiv.org/abs/1911.05722
    """

    def __init__(self, base_encoder, dim=128, K=65536, m=0.999, T=0.07, mlp=False):
        """
        dim: feature dimension (default: 128)
        K: queue size; number of negative keys (default: 65536)
        m: moco momentum of updating key encoder (default: 0.999)
        T: softmax temperature (default: 0.07)
        """
        super(MoCo, self).__init__()

        self.K = K
        self.m = m
        self.T = T

        # create the encoders
        # num_classes is the output fc dimension
        self.encoder_q = base_encoder(num_classes=dim)
        self.encoder_k = base_encoder(num_classes=dim)

        if mlp:  # hack: brute-force replacement
            dim_mlp = self.encoder_q.fc.weight.shape[1]
            self.encoder_q.fc = nn.Sequential(nn.Linear(dim_mlp, dim_mlp), nn.ReLU(), self.encoder_q.fc)
            self.encoder_k.fc = nn.Sequential(nn.Linear(dim_mlp, dim_mlp), nn.ReLU(), self.encoder_k.fc)

        for param_q, param_k in zip(self.encoder_q.parameters(), self.encoder_k.parameters()):
            param_k.data.copy_(param_q.data)  # initialize
            param_k.requires_grad = False  # not update by gradient

        # create the queue
        self.register_buffer("queue", torch.randn(dim, K))
        self.queue = nn.functional.normalize(self.queue, dim=0)

        self.register_buffer("queue_ptr", torch.zeros(1, dtype=torch.long))

    @torch.no_grad()
    def _momentum_update_key_encoder(self):
        """
        Momentum update of the key encoder
        """
        for param_q, param_k in zip(self.encoder_q.parameters(), self.encoder_k.parameters()):
            param_k.data = param_k.data * self.m + param_q.data * (1. - self.m)

    @torch.no_grad()
    def _dequeue_and_enqueue(self, keys):
        # gather keys before updating queue
        keys = concat_all_gather(keys)

        batch_size = keys.shape[0]

        ptr = int(self.queue_ptr)
        assert self.K % batch_size == 0  # for simplicity

        # replace the keys at ptr (dequeue and enqueue)
        self.queue[:, ptr:ptr + batch_size] = keys.T
        ptr = (ptr + batch_size) % self.K  # move pointer

        self.queue_ptr[0] = ptr

    @torch.no_grad()
    def _batch_shuffle_ddp(self, x):
        """
        Batch shuffle, for making use of BatchNorm.
        *** Only support DistributedDataParallel (DDP) model. ***
        """
        # gather from all gpus
        batch_size_this = x.shape[0]
        x_gather = concat_all_gather(x)
        batch_size_all = x_gather.shape[0]

        num_gpus = batch_size_all // batch_size_this

        # random shuffle index
        idx_shuffle = torch.randperm(batch_size_all).cuda()

        # broadcast to all gpus
        torch.distributed.broadcast(idx_shuffle, src=0)

        # index for restoring
        idx_unshuffle = torch.argsort(idx_shuffle)

        # shuffled index for this gpu
        gpu_idx = torch.distributed.get_rank()
        idx_this = idx_shuffle.view(num_gpus, -1)[gpu_idx]

        return x_gather[idx_this], idx_unshuffle

    @torch.no_grad()
    def _batch_unshuffle_ddp(self, x, idx_unshuffle):
        """
        Undo batch shuffle.
        *** Only support DistributedDataParallel (DDP) model. ***
        """
        # gather from all gpus
        batch_size_this = x.shape[0]
        x_gather = concat_all_gather(x)
        batch_size_all = x_gather.shape[0]

        num_gpus = batch_size_all // batch_size_this

        # restored index for this gpu
        gpu_idx = torch.distributed.get_rank()
        idx_this = idx_unshuffle.view(num_gpus, -1)[gpu_idx]

        return x_gather[idx_this]

    def forward(self, im_q, im_k):
        """
        Input:
            im_q: a batch of query images
            im_k: a batch of key images
        Output:
            logits, targets
        """

        # compute query features
        q = self.encoder_q(im_q)  # queries: NxC
        q = nn.functional.normalize(q, dim=1)

        # compute key features
        with torch.no_grad():  # no gradient to keys
            self._momentum_update_key_encoder()  # update the key encoder

            # shuffle for making use of BN
            im_k, idx_unshuffle = self._batch_shuffle_ddp(im_k)

            k = self.encoder_k(im_k)  # keys: NxC
            k = nn.functional.normalize(k, dim=1)

            # undo shuffle
            k = self._batch_unshuffle_ddp(k, idx_unshuffle)

        # compute logits
        # Einstein sum is more intuitive
        # positive logits: Nx1
        l_pos = torch.einsum('nc,nc->n', [q, k]).unsqueeze(-1)
        # negative logits: NxK
        l_neg = torch.einsum('nc,ck->nk', [q, self.queue.clone().detach()])

        # logits: Nx(1+K)
        logits = torch.cat([l_pos, l_neg], dim=1)

        # apply temperature
        logits /= self.T

        # labels: positive key indicators
        labels = torch.zeros(logits.shape[0], dtype=torch.long).cuda()

        # dequeue and enqueue
        self._dequeue_and_enqueue(k)

        return logits, labels


# utils
@torch.no_grad()
def concat_all_gather(tensor):
    """
    Performs all_gather operation on the provided tensors.
    *** Warning ***: torch.distributed.all_gather has no gradient.
    """
    tensors_gather = [torch.ones_like(tensor)
                      for _ in range(torch.distributed.get_world_size())]
    torch.distributed.all_gather(tensors_gather, tensor, async_op=False)

    output = torch.cat(tensors_gather, dim=0)
    return output



#!/usr/bin/env python3
import scriptconfig as scfg
import ubelt as ub


class CleanGeotiffConfig(scfg.DataConfig):
    r"""
    A preprocessing step for geotiff datasets.

    Replaces large contiguous regions of specific same-valued pixels as nodata.

    Note:
        This is a destructive operation and overwrites the geotiff image data
        inplace. Make a copy of your dataset if there is any chance you need to
        go back. The underlying kwcoco file is not modified.

    Usage:
        # It is a good idea to do a dry run first to check for issues
        # This can be done at a smaller scale for speed.
        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
        smartwatch clean_geotiffs \
            --src "$DVC_DATA_DPATH/Drop4-BAS/data.kwcoco.json" \
            --channels="red|green|blue|nir|swir16|swir22" \
            --prefilter_channels="red" \
            --min_region_size=256 \
            --nodata_value=-9999 \
            --workers="min(2,avail)" \
            --probe_scale=0.5 \
            --dry=True

        # Then execute a real run at full scale - optionally with a probe scale
        smartwatch clean_geotiffs \
            --src "$DVC_DATA_DPATH/Drop4-BAS/data_vali.kwcoco.json" \
            --channels="red|green|blue|nir|swir16|swir22" \
            --prefilter_channels="red" \
            --min_region_size=256 \
            --nodata_value=-9999 \
            --workers="min(2,avail)" \
            --probe_scale=None \
            --dry=False


    Ignore:

        smartwatch clean_geotiffs \
            --src=data.kwcoco.zip --dry=True --workers=8  \
            --probe_scale=0.25 --prefilter_channels=pan --channels=pan

        smartwatch clean_geotiffs \
            --dry=False --workers=2  \
            --probe_scale=0.0625 --prefilter_channels="pan" \
            --channels="pan" \
            --src=data.kwcoco.zip

        smartwatch clean_geotiffs \
            --dry=False --workers=avail  \
            --probe_scale=0.0625 --prefilter_channels="red" \
            --channels="red|green|blue|nir|swir16|swir22" \
            --src=data.kwcoco.zip

        smartwatch clean_geotiffs \
            --dry=False --workers=avail  \
            --probe_scale=0.25 --prefilter_channels="swir22" \
            --channels="swir16|swir22" \
            --src=imgonly-BR_R005.kwcoco.json

    """
    src = scfg.Value(None, position=1, help='input coco dataset')

    workers = scfg.Value(0, type=str, help='number of workers')

    channels = scfg.Value('red|green|blue|nir|swir16|swir22', help=ub.paragraph(
        '''
        The channels to apply nodata fixes to.
        '''))

    exclude_channels = scfg.Value('quality|cloudmask', help=ub.paragraph(
        '''
        Channels to never apply fixes to.
        '''))

    prefilter_channels = scfg.Value('red', help=ub.paragraph(
        '''
        The channels to use when checking for bad nodata values.
        If unspecified, uses channels.
        '''))

    possible_nodata_values = scfg.Value([0], help=ub.paragraph(
        '''
        List of integer values that might represent nodata, but may also failed
        to be labeled as nodata.
        '''))

    min_region_size = scfg.Value(256, help=ub.paragraph(
        '''
        Minimum size of a connected region to be considered as a nodata
        candidate.
        '''))

    scale = scfg.Value(None, help=ub.paragraph(
        '''
        Scale at which to perform the check. E.g. 0.5 for half resolution.
        Speeds up checks, but should only be used in dry runs.
        TODO: we could fix the real run to work with scale checks, and that
        would speed it up and probably not generate many false negatives.
        '''))

    nodata_value = scfg.Value(-9999, help='the real nodata value to use')

    probe_scale = scfg.Value(None, help=ub.paragraph(
        '''
        If specified, we probe the image at smaller scale for bad values and if
        any are found we do a full scale computation. Speeds up computation if
        there are not many images with bad values, but may result in false
        negatives.

        Should be given as a float less than 1.0
        '''))

    use_fix_stamps = scfg.Value(False, help=ub.paragraph(
        '''
        if True, write a file next to every file that was fixed so we dont need
        to check it again.
        '''))

    dry = scfg.Value(False, help='if True, only do a dry run. Report issues but do not fix them')


__config__ = CleanGeotiffConfig


def main(cmdline=1, **kwargs):
    """
    CommandLine:
        xdoctest -m watch.cli.coco_clean_geotiffs main

    Example:
        >>> # Generate a dataset that has bad nodata values
        >>> from watch.cli.coco_clean_geotiffs import *  # NOQA
        >>> import kwimage
        >>> import watch
        >>> import kwarray
        >>> import numpy as np
        >>> # Create a copy of the test dataset to clean inplace
        >>> orig_dset = watch.coerce_kwcoco('watch-msi', geodata=True, bad_nodata=True)
        >>> orig_dpath = ub.Path(orig_dset.bundle_dpath)
        >>> dpath = orig_dpath.augment(stemsuffix='_cleaned')
        >>> dpath.delete()
        >>> orig_dpath.copy(dpath)
        >>> dset = watch.coerce_kwcoco(dpath / 'data.kwcoco.json')
        >>> coco_img = dset.images().coco_images[0]
        >>> kwargs = {
        >>>     'src': dset,
        >>>     'workers': 0,
        >>>     'channels': 'B11',
        >>>     'prefilter_channels': 'B11',
        >>>     'min_region_size': 32,
        >>>     'nodata_value': 2,  # because toydata is uint16
        >>> }
        >>> cmdline = 0
        >>> # Do a dry run first
        >>> main(cmdline=cmdline, **kwargs, dry=True)
        >>> # Then a real run.
        >>> main(cmdline=cmdline, **kwargs)
        >>> coco_img1 = orig_dset.images().coco_images[0]
        >>> coco_img2 = dset.coco_image(coco_img1.img['id'])
        >>> print(ub.urepr(list(coco_img1.iter_image_filepaths())))
        >>> print(ub.urepr(list(coco_img2.iter_image_filepaths())))
        >>> imdata1 = coco_img1.imdelay('B11', nodata_method='float').finalize()
        >>> imdata2 = coco_img2.imdelay('B11', nodata_method='float').finalize()
        >>> print(np.isnan(imdata1).sum())
        >>> print(np.isnan(imdata2).sum())
        >>> canvas1 = kwarray.robust_normalize(imdata1)
        >>> canvas2 = kwarray.robust_normalize(imdata2)
        >>> canvas1 = kwimage.nodata_checkerboard(canvas1)
        >>> canvas2 = kwimage.nodata_checkerboard(canvas2)
        >>> # xdoctest: +REQUIRES(--show)
        >>> # xdoctest: +REQUIRES(module:kwplot)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(canvas1, pnum=(1, 2, 1), title='before')
        >>> kwplot.imshow(canvas2, pnum=(1, 2, 2), title='after')
    """
    from watch.utils import util_parallel
    import kwcoco

    config = CleanGeotiffConfig.cli(cmdline=cmdline, data=kwargs)
    import rich
    rich.print('config = {}'.format(ub.urepr(config, nl=1)))

    print('Loading dataset')
    coco_dset = kwcoco.CocoDataset.coerce(config['src'])

    workers = util_parallel.coerce_num_workers(config['workers'])
    print('workers = {}'.format(ub.urepr(workers, nl=1)))
    jobs = ub.JobPool(mode='process', max_workers=workers, transient=True)
    print('Created job pool')

    if config['channels'] is None or config['channels'] == '*':
        channels = None
    else:
        channels = kwcoco.ChannelSpec.coerce(config['channels'])
    if config['prefilter_channels'] is None:
        prefilter_channels = None
    else:
        prefilter_channels = kwcoco.ChannelSpec.coerce(config['prefilter_channels'])

    possible_nodata_values = set(config['possible_nodata_values'])

    if config['scale'] is not None:
        assert config['dry'], 'only scale in dry runs'

    probe_kwargs = {
        'channels': channels,
        'exclude_channels': kwcoco.ChannelSpec.coerce(config['exclude_channels']).fuse(),
        'scale': config['scale'],
        'possible_nodata_values': possible_nodata_values,
        'prefilter_channels': prefilter_channels,
        'min_region_size': config['min_region_size'],
        'probe_scale': config['probe_scale'],
        'use_fix_stamps': config['use_fix_stamps'],
        'nodata_value': config['nodata_value'],
    }

    print('Grabbing coco images')
    coco_imgs = coco_dset.images().coco_images
    print('About to start looping')

    from watch.utils import util_progress
    mprog = util_progress.ProgressManager(backend='progiter')
    # mprog = util_progress.ProgressManager(backend='rich')
    with mprog, jobs:
        print('In the with statement')
        mprog.update_info('Looking for geotiff issues')

        # TODO: may need to use the blocking job queue to limit the maximum
        # number of unhandled but populated results
        for coco_img in mprog.new(coco_imgs, desc='Submit probe jobs'):
            coco_img.detach()
            jobs.submit(probe_image_issues, coco_img, **probe_kwargs)

        def collect_jobs(jobs):
            for job in mprog.new(jobs.as_completed(), total=len(jobs), desc='Collect probe jobs'):
                image_summary = job.result()
                yield image_summary

        def filter_fixable(summaries):

            num_asset_issues = 0
            num_images_issues = 0
            num_incorrect_nodata = 0
            seen_bad_values = set()

            mprog.update_info(ub.codeblock(
                f'''
                Discovered Issues
                -----------------
                Found num_images_issues={num_images_issues}
                Found num_asset_issues={num_asset_issues}

                num_incorrect_nodata={num_incorrect_nodata}
                seen_bad_values={seen_bad_values}
                '''
            ))

            for image_summary in summaries:
                if image_summary['needs_fix']:
                    num_images_issues += 1
                    for asset_summary in image_summary['chans']:
                        if asset_summary['needs_fix']:
                            asset_summary['coco_img'] = image_summary['coco_img']
                            seen_bad_values.update(asset_summary['bad_values'])
                            if not asset_summary.get('has_incorrect_nodata_value', False):
                                num_incorrect_nodata += 1
                            num_asset_issues += 1
                            mprog.update_info(ub.codeblock(
                                f'''
                                Discovered Issues
                                -----------------
                                Found num_images_issues={num_images_issues}
                                Found num_asset_issues={num_asset_issues}

                                seen_bad_values={seen_bad_values}
                                '''
                            ))
                            yield asset_summary

        EAGER = 0

        summaries = collect_jobs(jobs)
        if EAGER:
            summaries = list(summaries)
        else:
            summaries = iter(summaries)

        needs_fix = filter_fixable(summaries)
        if EAGER:
            needs_fix = list(needs_fix)
        else:
            needs_fix = iter(needs_fix)

        if not config['dry']:
            correct_nodata_value = config['nodata_value']
            for asset_summary in mprog.new(needs_fix, desc='Cleaning identified issues'):
                fpath = ub.Path(asset_summary['fpath'])
                fix_geotiff_ondisk(asset_summary,
                                   correct_nodata_value=correct_nodata_value)
                if config['use_fix_stamps']:
                    # Mark that fixes have been applied to this asset
                    import json
                    fix_fpath = fpath.augment(tail='.fixes.stamp')
                    if fix_fpath.exists():
                        fixes = json.loads(fix_fpath.read_text())
                    else:
                        fixes = []
                    new_fixes = ub.udict(asset_summary) & {'band_idxs'}
                    fixes.append(new_fixes)
                    fix_fpath.write_text(json.dumps(fixes))
        else:
            for asset_summary in mprog.new(needs_fix, desc='Dry run: identifying issues'):
                print(asset_summary['fpath'])
                ...

    # if 0:
    #     import xdev
    #     import kwplot
    #     kwplot.autompl()

    #     nonzero_chans = []
    #     for image_summary in has_nonzeros:
    #         coco_img = image_summary['coco_img']
    #         for asset_summary in image_summary['chans']:
    #             bad_values = asset_summary['bad_values']
    #             if len(set(bad_values) - {0}):
    #                 asset_summary['coco_img'] = coco_img
    #                 nonzero_chans.append(asset_summary)

    #     iter_ = xdev.InteractiveIter(nonzero_chans)
    #     for asset_summary in iter_:
    #         coco_img = asset_summary['coco_img']
    #         chan_canvas = draw_asset_summary(coco_img, asset_summary)
    #         kwplot.imshow(chan_canvas, doclf=1)
    #         xdev.InteractiveIter.draw()


def probe_image_issues(coco_img, channels=None, prefilter_channels=None, scale=None,
                       possible_nodata_values=None, min_region_size=256,
                       exclude_channels=None, probe_scale=None,
                       use_fix_stamps=False, nodata_value=-9999):
    """
    Inspect a single image, possibily with multiple assets, each with possibily
    multiple bands for fixable nodata values.

    Args:
        coco_img : the coco image to check

        channels : the channels to check

        prefilter_channels : the channels to check first for efficiency

        possible_nodata_values (set[int]):
            the values that may be nodata if known

        scale :
            use a downscaled overview to speed up the computation via
            approximation. Returns results at this scale. DO NOT USE RIGHT NOW.

        probe_scale :
            use a downscaled overview to speed up the computation via
            approximation. If the probe identifies an issue a full scale probe
            is done.

    Ignore:
        globals().update((ub.udict(xdev.get_func_kwargs(probe_image_issues)) | probe_kwargs))

    Example:
        >>> import watch
        >>> from watch.cli.coco_clean_geotiffs import *  # NOQA
        >>> import numpy as np
        >>> dset = watch.coerce_kwcoco('watch-msi', geodata=True, bad_nodata=True)
        >>> coco_img = dset.images().coco_images[4]
        >>> channels = 'B11|B10|X.1'
        >>> prefilter_channels = 'B11'
        >>> scale = None
        >>> possible_nodata_values = {0}
        >>> min_region_size = 128
        >>> image_summary = probe_image_issues(
        >>>     coco_img, channels=channels, prefilter_channels=prefilter_channels,
        >>>     scale=scale, possible_nodata_values=possible_nodata_values,
        >>>     min_region_size=min_region_size)
        >>> print(f'image_summary={image_summary}')
    """
    import kwcoco
    if channels is None:
        request_channels = coco_img.channels.fuse()
    else:
        request_channels = kwcoco.ChannelSpec.coerce(channels).fuse()
    if prefilter_channels is not None:
        prefilter_channels = kwcoco.ChannelSpec.coerce(prefilter_channels).fuse()

    if exclude_channels is not None:
        request_channels = request_channels.difference(exclude_channels)
        if prefilter_channels is not None:
            prefilter_channels = prefilter_channels.difference(exclude_channels)

    image_summary = {}
    asset_summaries = []

    prefilter_assets = []
    requested_assets = []
    for obj in coco_img.iter_asset_objs():
        chans = kwcoco.ChannelSpec.coerce(obj['channels']).fuse()
        request_common = (chans & request_channels).fuse()
        if request_common.numel():
            if prefilter_channels is None:
                prefilter_common = None
            else:
                prefilter_common = (chans & prefilter_channels).fuse()
            chans_ = chans.to_oset()
            band_idxs = [chans_.index(c) for c in request_common.to_list()]
            if prefilter_common is not None and prefilter_common.numel():
                prefilter_assets.append({
                    'obj': obj,
                    'band_idxs': band_idxs,
                })
            else:
                requested_assets.append({
                    'obj': obj,
                    'band_idxs': band_idxs,
                })

    should_continue = True
    for item in prefilter_assets:
        obj = item['obj']
        band_idxs = item['band_idxs']

        if probe_scale is not None:
            # Check if there is anything obvious at a smaller scale first if
            # requested.
            probe_asset_summary = probe_asset(
                coco_img, obj, band_idxs=band_idxs, scale=probe_scale,
                possible_nodata_values=possible_nodata_values,
                min_region_size=min_region_size, use_fix_stamps=use_fix_stamps,
                nodata_value=nodata_value)
            if not probe_asset_summary['bad_values']:
                should_continue = False

        if should_continue:
            # Check if there is anything obvious in a specific channel first if
            # requested.
            asset_summary = probe_asset(coco_img, obj, band_idxs=band_idxs,
                                        scale=scale,
                                        possible_nodata_values=possible_nodata_values,
                                        min_region_size=min_region_size,
                                        use_fix_stamps=use_fix_stamps,
                                        nodata_value=nodata_value)
            if not asset_summary['bad_values']:
                should_continue = False
            asset_summaries.append(asset_summary)

    if len(prefilter_assets) == 0:
        should_continue = True

    if should_continue:
        for item in requested_assets:
            obj = item['obj']
            band_idxs = item['band_idxs']
            asset_summary = probe_asset(coco_img, obj, band_idxs=band_idxs,
                                        scale=scale,
                                        possible_nodata_values=possible_nodata_values,
                                        min_region_size=min_region_size,
                                        use_fix_stamps=use_fix_stamps,
                                        nodata_value=nodata_value)
            asset_summaries.append(asset_summary)

    image_summary['chans'] = asset_summaries
    image_summary['coco_img'] = coco_img
    image_summary['bad_values'] = list(ub.unique(ub.flatten([c['bad_values'] for c in asset_summaries])))
    image_summary['needs_fix'] = any([c['needs_fix'] for c in asset_summaries])
    return image_summary


def probe_asset(coco_img, obj, band_idxs=None, scale=None,
                possible_nodata_values=None, min_region_size=256,
                use_fix_stamps=False, nodata_value=-9999):
    """
    Inspect a specific single-file asset possibily with multiple bands for
    fixable nodata values.
    """
    import kwcoco
    asset_channels = kwcoco.FusedChannelSpec.coerce(obj['channels'])
    if band_idxs is None:
        band_idxs = list(range(asset_channels.numel()))

    bundle_dpath = ub.Path(coco_img.bundle_dpath)
    fpath = bundle_dpath / obj['file_name']

    asset_summary = {
        'channels': asset_channels,
        'fpath': fpath,
        'needs_fix': False,
    }

    if use_fix_stamps:
        import json
        fix_fpath = fpath.augment(tail='.fixes.stamp')
        if fix_fpath.exists():
            fixes = json.loads(fix_fpath.read_text())
            assert fixes
            return asset_summary
            # TODO: actually check the requested fix was applied, but for now
            # assume if the file exists we should skip the check.

    check_channels = asset_channels[band_idxs]
    delayed = coco_img.imdelay(
        channels=check_channels, space='asset', nodata_method='float')
    if scale is not None:
        delayed.prepare()
        delayed = delayed.scale(scale)
        delayed = delayed.optimize()

    imdata = delayed.finalize(interpolation='nearest')

    if scale is not None:
        min_region_size_ = int(min_region_size * scale)
    else:
        min_region_size_ = min_region_size

    if possible_nodata_values is not None:
        possible_nodata_values = list(possible_nodata_values)
        possible_nodata_values.append(nodata_value)

    asset_summary.update(probe_asset_imdata(
        imdata, band_idxs, min_region_size_=min_region_size_,
        possible_nodata_values=possible_nodata_values))

    asset_summary.update(_probe_correct_nodata_value(
        fpath, band_idxs, nodata_value))

    return asset_summary


def _probe_correct_nodata_value(fpath, band_idxs, nodata_value=-9999):
    asset_summary = {}
    from watch.utils import util_gdal
    gdal_dset = util_gdal.GdalDataset.open(fpath)
    band_infos = gdal_dset.info()['bands']
    gdal_dset = None
    asset_summary['has_incorrect_nodata_value'] = False
    for band_idx in band_idxs:
        band_info = band_infos[band_idx]
        if band_info.get('noDataValue', None) != nodata_value:
            asset_summary['has_incorrect_nodata_value'] = True
            asset_summary['needs_fix'] = True
    return asset_summary


def probe_asset_imdata(imdata, band_idxs, min_region_size_=256,
                       possible_nodata_values=None):
    import kwarray
    asset_summary = {}
    asset_summary['band_idxs'] = band_idxs
    band_summaries = []
    imdata = kwarray.atleast_nd(imdata, n=3)
    for idx in range(len(band_idxs)):
        band_imdata = imdata[..., idx]
        band_summary = probe_band_imdata(
            band_imdata, min_region_size_=min_region_size_,
            possible_nodata_values=possible_nodata_values)
        band_summaries.append(band_summary)
    asset_summary['band_summaries'] = band_summaries
    asset_summary['bad_values'] = list(ub.unique(ub.flatten([
        c['bad_values'] for c in band_summaries])))
    if len(asset_summary['bad_values']) > 0:
        asset_summary['needs_fix'] = True
    return asset_summary


def probe_band_imdata(band_imdata, min_region_size_=256,
                      possible_nodata_values=None):
    from watch.utils import util_kwimage
    import numpy as np

    band_imdata = np.ascontiguousarray(band_imdata)
    is_samecolor = util_kwimage.find_samecolor_regions(
        band_imdata, min_region_size=min_region_size_,
        values=possible_nodata_values, PRINT_STEPS=0)

    band_summary = {}
    if np.any(is_samecolor):
        unique_labels, unique_first_index, unique_counts = np.unique(is_samecolor, return_index=True, return_counts=True)
        flags = unique_labels > 0
        bad_labels = unique_labels[flags]
        first_index = unique_first_index[flags]
        bad_counts = unique_counts[flags]
        bad_values = band_imdata.ravel()[first_index]

        band_summary['is_samecolor'] = is_samecolor
        band_summary['bad_labels'] = bad_labels
        band_summary['bad_values'] = bad_values
        band_summary['bad_counts'] = bad_counts
    else:
        band_summary['is_samecolor'] = False
        band_summary['bad_values'] = []
    return band_summary


def fix_single_asset(fpath, dry=False):
    from delayed_image import DelayedLoad
    asset_summary = {
        'fpath': fpath,
        'needs_fix': False,
    }

    min_region_size_ = 256
    nodata_value = -9999
    possible_nodata_values = [0, nodata_value]

    delayed = DelayedLoad(fpath)
    delayed.prepare()
    band_idxs = list(range(delayed.num_channels))

    imdata = delayed.finalize(nodata_method='nan', interpolation='nearest')

    asset_summary.update(probe_asset_imdata(
        imdata, band_idxs, min_region_size_=min_region_size_,
        possible_nodata_values=possible_nodata_values))

    asset_summary.update(_probe_correct_nodata_value(
        fpath, band_idxs, nodata_value))

    print('asset_summary = {}'.format(ub.urepr(asset_summary, nl=True)))
    if not dry:
        fix_geotiff_ondisk(asset_summary, correct_nodata_value=nodata_value)


def fix_geotiff_ondisk(asset_summary, correct_nodata_value=-9999):
    """
    Updates the nodata value based on a mask inplace on disk.
    Attempts to preserve all other metadata, but this is not guarenteed or
    always possible.

    Args:
        asset_summary (Dict): an item from :func:`probe_asset`.

        correct_nodata_value (int): the nodata value to use in the
            modified geotiff.

    Assumptions:
        * The input image uses AVERAGE overview resampling
        * The input image is a tiled geotiff (ideally a COG)

    TODO:
        - [ ] Can restructure this as a more general context manager.

    Example:
        >>> from watch.cli.coco_clean_geotiffs import *  # NOQA
        >>> from watch.demo.metrics_demo.demo_rendering import write_demo_geotiff
        >>> import kwimage
        >>> import numpy as np
        >>> dpath = ub.Path.appdir('watch/tests/clean_geotiff').ensuredir()
        >>> fpath1 = dpath / 'test_geotiff.tif'
        >>> fpath2 = fpath1.augment(stemsuffix='_fixed')
        >>> fpath1.delete()
        >>> fpath2.delete()
        >>> imdata = kwimage.grab_test_image('amazon', dsize=(512, 512))
        >>> poly = kwimage.Polygon.random().scale(imdata.shape[0:2][::-1])
        >>> imdata = poly.draw_on(imdata, color='black')
        >>> imdata = imdata.astype(np.int16)
        >>> #imdata = poly.fill(imdata, value=(0, 0, 0), pixels_are='areas')
        >>> imdata = poly.fill(imdata, value=0, pixels_are='areas')
        >>> imdata[:256, :256, 0] = 0
        >>> write_demo_geotiff(img_fpath=fpath1, imdata=imdata)
        >>> fpath1.copy(fpath2)
        >>> asset_summary = probe_asset_imdata(imdata, band_idxs=[0, 2], possible_nodata_values={0})
        >>> asset_summary['fpath'] = fpath2
        >>> assert fpath1.stat().st_size == fpath2.stat().st_size
        >>> fix_geotiff_ondisk(asset_summary)
        >>> assert fpath1.stat().st_size != fpath2.stat().st_size
        >>> imdata1 = kwimage.imread(fpath1, nodata_method='ma')
        >>> imdata2 = kwimage.imread(fpath2, nodata_method='ma')
        >>> canvas1 = kwimage.normalize_intensity(imdata1)
        >>> canvas2 = kwimage.normalize_intensity(imdata2)
        >>> canvas1 = kwimage.nodata_checkerboard(canvas1)
        >>> canvas2 = kwimage.nodata_checkerboard(canvas2)
        >>> # xdoctest: +REQUIRES(--show)
        >>> # xdoctest: +REQUIRES(module:kwplot)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(canvas1.data, pnum=(2, 2, 1), title='norm imdata1 vals')
        >>> kwplot.imshow(canvas2.data, pnum=(2, 2, 2), title='norm imdata2 vals')
        >>> kwplot.imshow(imdata1.mask.any(axis=2), pnum=(2, 2, 3), title='imdata1.mask')
        >>> kwplot.imshow(imdata2.mask.any(axis=2), pnum=(2, 2, 4), title='imdata2.mask')
        >>> #kwplot.imshow((asset_summary['is_samecolor'] > 0), pnum=(3, 2, 5), title='is samecolor mask')

    Example:
        >>> from watch.cli.coco_clean_geotiffs import *  # NOQA
        >>> from watch.demo.metrics_demo.demo_rendering import write_demo_geotiff
        >>> import kwimage
        >>> import numpy as np
        >>> dpath = ub.Path.appdir('watch/tests/clean_geotiff').ensuredir()
        >>> fpath1 = dpath / 'test_geotiff.tif'
        >>> fpath2 = fpath1.augment(stemsuffix='_fixed')
        >>> fpath1.delete()
        >>> fpath2.delete()
        >>> imdata = kwimage.grab_test_image('amazon', dsize=(512, 512))[..., 0]
        >>> poly = kwimage.Polygon.random().scale(imdata.shape[0:2][::-1])
        >>> imdata = imdata.astype(np.int16)
        >>> imdata = poly.fill(imdata, value=0, pixels_are='areas')
        >>> imdata[:256, :256] = 0
        >>> write_demo_geotiff(img_fpath=fpath1, imdata=imdata)
        >>> fpath1.copy(fpath2)
        >>> asset_summary = probe_asset_imdata(imdata, band_idxs=[0], possible_nodata_values={0})
        >>> asset_summary['fpath'] = fpath2
        >>> assert fpath1.stat().st_size == fpath2.stat().st_size
        >>> fix_geotiff_ondisk(asset_summary)
        >>> assert fpath1.stat().st_size != fpath2.stat().st_size
        >>> imdata1 = kwimage.imread(fpath1, nodata_method='ma')
        >>> imdata2 = kwimage.imread(fpath2, nodata_method='ma')
        >>> canvas1 = kwimage.normalize_intensity(imdata1)
        >>> canvas2 = kwimage.normalize_intensity(imdata2)
        >>> canvas1 = kwimage.nodata_checkerboard(canvas1)
        >>> canvas2 = kwimage.nodata_checkerboard(canvas2)
        >>> # xdoctest: +REQUIRES(--show)
        >>> # xdoctest: +REQUIRES(module:kwplot)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(canvas1.data, pnum=(2, 2, 1), title='norm imdata1 vals')
        >>> kwplot.imshow(canvas2.data, pnum=(2, 2, 2), title='norm imdata2 vals')
        >>> kwplot.imshow(imdata1.mask, pnum=(2, 2, 3), title='imdata1.mask')
        >>> kwplot.imshow(imdata2.mask, pnum=(2, 2, 4), title='imdata2.mask')
        >>> #kwplot.imshow((asset_summary['is_samecolor'] > 0), pnum=(3, 2, 5), title='is samecolor mask')
    """
    import os
    from watch.utils import util_gdal
    from osgeo import gdal
    import numpy as np
    import tempfile
    fpath = ub.Path(asset_summary['fpath'])
    band_idxs = asset_summary['band_idxs']

    # We will write a modified version to a temporay file and then overwrite
    # the destination file to avoid race conditions.
    tmp_fpath = ub.Path(tempfile.mktemp(
        dir=fpath.parent, prefix='.tmp.' + fpath.stem, suffix='.tiff'))
    assert tmp_fpath.parent.exists()

    # Assume average overview resampling, there does not seem to be standard
    # way that a geotiff encodes what algorithm was used.
    overview_resample = 'AVERAGE'

    ### Read in source data and introspect it
    src_dset = util_gdal.GdalDataset.open(os.fspath(fpath))
    orig_driver_name = src_dset.GetDriver().ShortName
    assert orig_driver_name == 'GTiff'

    band_idx_to_orig_props = {}
    for band_idx in band_idxs:
        band = src_dset.GetRasterBand(band_idx + 1)
        # Have to read this here because it is clobbered in the copy
        band_idx_to_orig_props[band_idx] = {
            'blocksize': band.GetBlockSize(),
            'num_overviews': band.GetOverviewCount(),
        }

    ### Copy the source data to memory and modify it
    driver1 = gdal.GetDriverByName(str('MEM'))
    copy1 = driver1.CreateCopy(str(''), src_dset)
    src_dset.FlushCache()
    src_dset = None

    ### Pixel Modification ###
    # Modify the pixel contents
    for band_idx, band_summary in zip(band_idxs, asset_summary['band_summaries']):
        band = copy1.GetRasterBand(band_idx + 1)
        band_data = band.ReadAsArray()
        mask = band_summary['is_samecolor'] > 0

        curr_nodat_value = band.GetNoDataValue()

        if curr_nodat_value is not None:
            if curr_nodat_value != correct_nodata_value:
                is_curr_nodata = (band_data == curr_nodat_value)
                mask = (mask | is_curr_nodata)

        band_data[mask] = correct_nodata_value

        band.WriteArray(band_data)
        if curr_nodat_value != correct_nodata_value:
            band.SetNoDataValue(correct_nodata_value)

    if band_idx_to_orig_props:
        assert ub.allsame(band_idx_to_orig_props.values()), (
            'We expect input bands to have the same blocksize')
        bandprop = ub.peek(band_idx_to_orig_props.values())
        blocksize = bandprop['blocksize']
        num_overviews = bandprop['num_overviews']
    else:
        raise AssertionError('should have this info')

    ### Rebuild overviews
    overviewlist = (2 ** np.arange(1, num_overviews + 1)).tolist()
    copy1.BuildOverviews(overview_resample, overviewlist)

    ### Flush the in-memory dataset to an on-disk GeoTiff
    xsize, ysize = blocksize
    _options = [
        'BIGTIFF=YES',
        'TILED=YES',
        'BLOCKXSIZE={}'.format(xsize),
        'BLOCKYSIZE={}'.format(ysize),
    ]
    _options += ['COMPRESS={}'.format('DEFLATE')]
    _options.append('COPY_SRC_OVERVIEWS=YES')
    driver1 = None
    # driver2 = gdal.GetDriverByName(str('GTiff'))
    driver2 = gdal.GetDriverByName(str(orig_driver_name))
    copy2 = driver2.CreateCopy(os.fspath(tmp_fpath), copy1, options=_options)

    if copy2 is None:
        last_gdal_error = gdal.GetLastErrorMsg()
        if 'No such file or directory' in last_gdal_error:
            ex_cls = IOError
        else:
            ex_cls = Exception
        raise ex_cls(
            'Unable to create gtiff driver for fpath={}, options={}, last_gdal_error={}'.format(
                fpath, _options, last_gdal_error))
    copy2.FlushCache()
    copy1 = None
    copy2 = None  # NOQA
    driver2 = None

    assert ub.Path(tmp_fpath).exists()
    os.rename(tmp_fpath, fpath)


def draw_asset_summary(coco_img, asset_summary):
    from watch.utils import util_kwimage
    import kwimage
    import numpy as np
    is_samecolor = asset_summary['is_samecolor']
    data = coco_img.imdelay(channels=asset_summary['channels'], space='asset',
                            nodata_method='float').finalize()

    is_samecolor = np.ascontiguousarray(is_samecolor)
    unique_labels, first_index = np.unique(is_samecolor, return_index=True)
    data_values = data.ravel()[first_index]

    # is_samecolor.ravel()[first_index]
    label_mapping = dict(zip(unique_labels, data_values))
    label_mapping[0] = 'valid data'
    labels = is_samecolor
    label_to_color = {0: 'black'}
    label_canvas = util_kwimage.colorize_label_image(
        labels, with_legend=True,
        label_mapping=label_mapping, label_to_color=label_to_color)

    poly = kwimage.Mask.coerce(is_samecolor > 0).to_multi_polygon()
    boxes = kwimage.Boxes.concatenate([p.to_boxes() for p in poly.data])

    im_canvas = kwimage.normalize_intensity(data)
    im_canvas = kwimage.fill_nans_with_checkers(im_canvas, on_value=0.3)
    im_canvas = boxes.draw_on(im_canvas)
    canvas = kwimage.stack_images([im_canvas, label_canvas], axis=1)
    title = coco_img.video['name'] + ' ' + asset_summary['channels'] + '\n' + coco_img.img['name']
    canvas = kwimage.draw_header_text(canvas, title)

    if 0:
        import kwplot
        kwplot.imshow(canvas)
    return canvas


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/watch/watch/cli/coco_clean_geotiffs.py
    """
    main()
